from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Mapping,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
)

from typing_extensions import AbstractSet

DependencyLabel = Optional[str]
DependencySetup = Callable[..., None]
DependencyAsyncSetup = Callable[..., Coroutine[None, None, None]]

RateProgress = Callable[[float], Awaitable[None]]
# TODO: remove this when breaking API
PercentProgress = RateProgress
RawProgress = Callable[[int], Awaitable[None]]

DictStrAny = Dict[str, Any]
IntStr = Union[int, str]
AbstractSetIntStr = AbstractSet[IntStr]
MappingIntStrAny = Mapping[IntStr, Any]


class DependencyTeardown(Protocol):
    def __call__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> None: ...


class DependencyAsyncTeardown(Protocol):
    async def __call__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> None: ...


Dependency = Tuple[
    DependencyLabel,
    Union[DependencySetup, DependencyAsyncSetup],
    Optional[Union[DependencyTeardown, DependencyAsyncTeardown]],
]



import os
from abc import ABC
from copy import deepcopy
from typing import ClassVar, Generic, TypeVar

from pydantic import BaseSettings, validator
from pydantic.fields import FieldInfo
from pydantic.generics import GenericModel

from icij_common.pydantic_utils import safe_copy
from icij_worker.task_manager import TaskManagerConfig
from icij_worker.utils.registrable import find_variable_loc_in_env

TM = TypeVar("TM", bound=TaskManagerConfig)  # pylint: disable=invalid-name


class SettingsWithTM(BaseSettings, GenericModel, Generic[TM], ABC):
    """Helper which lets you configure your app which wraps a task manager and read the
     app settings from env like this:

    MY_OTHER_APP_SETTING=ANOTHER_SETTING
    TASK_MANAGER__APP=icij_worker.utils.tests.APP
    TASK_MANAGER__BACKEND=amqp
    TASK_MANAGER__STORAGE__MAX_CONNECTIONS=28
    TASK_MANAGER__RABBITMQ_HOST=localhost
    TASK_MANAGER__RABBITMQ_PORT=15752
    """

    app_path: ClassVar[str]

    task_manager: TM

    # Your other setting go here
    # my_other_setting: str

    class Config:  # Override this as needed
        env_prefix = ""
        env_nested_delimiter = "__"

    @classmethod
    def from_env(cls):
        tm_field_name = cls.__fields__["task_manager"].name
        backend_field_name = TaskManagerConfig.registry_key.default
        tm_backend_env_key = (
            f"{cls.__config__.env_prefix}{tm_field_name}"
            f"{cls.__config__.env_nested_delimiter}{backend_field_name}"
        )
        backend_key, backend = find_variable_loc_in_env(
            tm_backend_env_key, cls.__config__.case_sensitive
        )
        tm_config_cls = TaskManagerConfig.resolve_class_name(backend)
        old_env = deepcopy(os.environ)
        try:
            os.environ.pop(backend_key)
            return cls[tm_config_cls]()
        finally:
            os.environ.clear()
            os.environ.update(old_env)

    @validator("task_manager", pre=True, always=True)
    def set_task_manager_app_path(cls, v: TM):
        # pylint: disable=no-self-argument
        app_path = cls.app_path
        if isinstance(app_path, FieldInfo):
            app_path = app_path.default
        if isinstance(v, dict):
            v["app_path"] = app_path
        elif isinstance(v, TaskManagerConfig):
            v = safe_copy(v, update={"app_path": app_path})
        return v

    def to_task_manager(self) -> "TaskManager":
        from icij_worker import TaskManager

        return TaskManager.from_config(self.task_manager)



from .app import AsyncApp
from .objects import (
    Task,
    TaskError,
    ManagerEvent,
    ResultEvent,
    TaskState,
    Message,
    AsyncBackend,
)
from .task_manager import TaskManager
from .worker import Worker, WorkerConfig
from .routing_strategy import RoutingStrategy

try:
    from icij_worker.worker.amqp import AMQPWorker, AMQPWorkerConfig
    from icij_worker.event_publisher.amqp import AMQPPublisher
    from icij_worker.task_manager.amqp import AMQPTaskManager, AMQPTaskManagerConfig
except ImportError:
    pass

try:
    from icij_worker.worker.neo4j_ import Neo4jWorker, Neo4jWorkerConfig
    from icij_worker.event_publisher.neo4j_ import Neo4jEventPublisher
    from icij_worker.task_manager.neo4j_ import Neo4JTaskManager, Neo4JTaskManagerConfig
except ImportError:
    pass

try:
    from icij_worker.task_storage.fs import FSKeyValueStorage, FSKeyValueStorageConfig
except ImportError:
    pass

try:
    from icij_worker.task_storage.postgres import (
        PostgresStorage,
        PostgresStorageConfig,
        PostgresConnectionInfo,
        init_database as init_postgres_database,
    )
except ImportError:
    pass

from .backend import WorkerBackend
from .event_publisher import EventPublisher



from __future__ import annotations

import functools
from contextlib import AsyncExitStack
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional, Type, Union

import ujson
from sqlitedict import SqliteDict

from icij_common.pydantic_utils import ICIJModel
from icij_worker import Task, TaskState
from icij_worker.exceptions import UnknownTask
from icij_worker.task_storage import TaskStorageConfig
from icij_worker.task_storage.key_value import DBItem, KeyValueStorage


class FSKeyValueStorageConfig(ICIJModel, TaskStorageConfig):
    db_path: Path

    def to_storage(self) -> FSKeyValueStorage:
        return FSKeyValueStorage(self.db_path)


class FSKeyValueStorage(KeyValueStorage):
    # Save each type in a different DB to speedup lookup
    _tasks_db_name = "tasks"
    _results_db_name = "results"
    _errors_db_name = "errors"
    # pylint: disable=c-extension-no-member
    _encode = functools.partial(ujson.encode, default=str)
    _decode = functools.partial(ujson.decode)

    # pylint: enable=c-extension-no-member

    def __init__(self, db_path: Path):
        self._db_path = str(db_path)
        self._exit_stack = AsyncExitStack()
        self._dbs = dict()
        # TODO: add support for 1 DB / group
        self._dbs = self._make_group_dbs()

    async def __aenter__(self):
        for db in self._dbs.values():
            self._exit_stack.enter_context(db)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

    async def _read_key(self, db: str, *, key: str):
        db = self._dbs[db]
        try:
            return db[key]
        except KeyError as e:
            raise UnknownTask(key) from e

    async def _insert(self, db: str, obj: DBItem, *, key: str):
        db = self._dbs[db]
        try:
            db[key] = obj
        except KeyError as e:
            raise UnknownTask(key) from e
        db.commit()

    async def _update(self, db: str, update: DBItem, *, key: str):
        db = self._dbs[db]
        try:
            task = db[key]
        except KeyError as e:
            raise UnknownTask(key) from e
        task.update(update)
        db[key] = task
        db.commit()

    async def _add_to_array(self, db: str, obj: DBItem, *, key: str):
        db = self._dbs[db]
        values = deepcopy(db.get(key, []))
        values.append(obj)
        db[key] = values
        db.commit()

    def _key(self, task_id: str, obj_cls: Type) -> str:
        # Since each object type is saved in a different DB, we index by task ID
        return task_id

    async def get_tasks(
        self,
        group: Optional[str],
        *,
        task_name: Optional[str] = None,
        state: Optional[Union[List[TaskState], TaskState]] = None,
        **kwargs,
    ) -> List[Task]:
        states = set()
        if state is not None:
            if isinstance(state, TaskState):
                states = {state}
            states = set(states)
        tasks = self._dbs[self._tasks_db_name].values()
        if group is not None:
            tasks = (t for t in tasks if t.get("group") == group)
        if task_name is not None:
            tasks = (t for t in tasks if t["name"] == task_name)
        if states:
            tasks = (t for t in tasks if t["state"] in states)
        tasks = list(tasks)
        for t in tasks:
            t.pop("group", None)
        task = [Task.parse_obj(t) for t in tasks]
        return task

    def _make_db(self, filename: str, *, name: str) -> SqliteDict:
        return SqliteDict(
            filename,
            tablename=name,
            encode=self._encode,
            decode=self._decode,
            journal_mode="DEFAULT",
        )

    def _make_group_dbs(self) -> Dict[str, SqliteDict]:
        return {
            self._tasks_db_name: self._make_db(self._db_path, name=self._tasks_db_name),
            self._results_db_name: self._make_db(
                self._db_path, name=self._results_db_name
            ),
            self._errors_db_name: self._make_db(
                self._db_path, name=self._errors_db_name
            ),
        }
